#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import streamlit as st

st.set_page_config(
    page_title="Pricing xd",
    page_icon="💰",
    layout="wide",
)

chatbot_help_msg = """
## Pricing 💰

*Pricing data from [OpenAI](https://openai.com/pricing) and [Anthropic](https://docs.anthropic.com/claude/docs/models-overview#model-comparison)*

| Model Name                 | Pricing for 1M Tokens (Input/Output) (USD) | Cost for 1 Hour Audio (USD) |
|----------------------------|-------------------------------------------------|----------------------------------|
| `gpt-3.5-turbo-0125`       | 0.5, 1.5                                        | 0.01                             |
| `gpt-3.5-turbo`            | 0.5, 1.5                                        | 0.01                             |
| `gpt-4-0125-preview`       | 10, 30                                          | 0.5                              |
| `gpt-4-turbo-preview`      | 10, 30                                          | 0.5                              |
| `claude-3-haiku-20240307`  | 0.25, 1.25                                      | 0.015                            |
| `claude-3-sonnet-20240229` | 3, 15                                           | 0.2                              |
| `claude-3-opus-20240229`   | 15, 75                                          | 1                                |

**Note: The cost is estimated based on the token count of the input and output text. The actual cost may vary due to the language and audio speed.**

### Recommended Translation Model

- For English audio, we recommend using `gpt-3.5-turbo`.
- For non-English audio, we recommend using `claude-3-sonnet-20240229`.
"""
st.markdown(chatbot_help_msg, unsafe_allow_html=True)



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import json
import re
import sys
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import List, Union, Dict

from openlrc.logger import logger
from openlrc.utils import format_timestamp, parse_timestamp, detect_lang


@dataclass
class Element:
    """
    Save a LRC format element.
    """
    start: float
    end: Union[float, None]
    text: str

    @property
    def duration(self):
        if self.end:
            return self.end - self.start
        else:
            return sys.maxsize  # Fake int infinity

    def to_json(self):
        return {'start': self.start, 'end': self.end, 'text': self.text}


class Subtitle:
    """
    Save a sequence of Element, and meta data.
    """

    def __init__(self, language: str, segments: List[Dict], filename: Union[str, Path]):
        self.lang = language
        self.segments: List[Element] = [Element(**seg) for seg in segments]
        self.filename = Path(filename)

    @staticmethod
    def from_json(filename):
        with open(filename, encoding='utf-8') as f:
            content = json.loads(f.read())
        return Subtitle(filename=filename, **content)

    @staticmethod
    def from_file(filename):
        filename = Path(filename)
        suffix = filename.suffix
        if suffix == '.json':
            return Subtitle.from_json(filename)
        elif suffix == '.lrc':
            return Subtitle.from_lrc(filename)
        elif suffix == '.srt':
            return Subtitle.from_srt(filename)

    def __len__(self):
        return len(self.segments)

    @property
    def texts(self):
        return [e.text for e in self.segments]

    def set_texts(self, texts, lang=None):
        # Check length
        assert len(texts) == len(self.segments)

        for i, text in enumerate(texts):
            self.segments[i].text = text

        if lang:
            self.lang = lang

    def save(self, filename: Union[str, Path], update_name=False):
        results = {
            'language': self.lang,
            'segments': [seg.to_json() for seg in self.segments]
        }

        with open(filename, 'w', encoding='utf-8') as f:
            json.dump(results, f, ensure_ascii=False, indent=4)

        if update_name:
            self.filename = filename

        return filename

    def exists(self):
        """
        Check if the .lrc/.srt/.json file exist.
        :return:
        """
        lrc_path = self.filename.with_suffix('.lrc')
        srt_path = self.filename.with_suffix('.srt')

        return lrc_path.exists() or srt_path.exists() or self.filename.exists()

    def to_lrc(self):
        # If duration larger than 1 hour, use srt file instead
        if self.segments[-1].end >= 3600:
            logger.warning('Duration larger than 1 hour, use srt file instead')
            self.to_srt()
            return self.filename.with_suffix('.srt')

        lrc_path = self.filename.with_suffix('.lrc')
        fmt = partial(format_timestamp, fmt='lrc')
        with open(lrc_path, 'w', encoding='utf-8') as f:
            for i, segment in enumerate(self.segments):
                print(
                    f'[{fmt(segment.start)}] {segment.text}',
                    file=f,
                    flush=True,
                )
                if i == len(self.segments) - 1 or segment.end != self.segments[i + 1].start:
                    print(f'[{fmt(segment.end)}]', file=f, flush=True)

        logger.info(f'File saved to {lrc_path}')

        return lrc_path

    def to_srt(self):
        srt_path = self.filename.with_suffix('.srt')
        fmt = partial(format_timestamp, fmt='srt')
        with open(srt_path, 'w', encoding='utf-8') as f:
            for i, segment in enumerate(self.segments, start=1):
                print(f'{i}\n'
                      f'{fmt(segment.start)} --> {fmt(segment.end)}\n'
                      f'{segment.text}\n', file=f, flush=True)

        logger.info(f'File saved to {srt_path}')

        return srt_path

    @classmethod
    def from_lrc(cls, filename):
        filename = Path(filename)
        with open(filename, encoding='utf-8') as f:
            lines = f.readlines()

        # Remove comments
        lines = [line for line in lines if not line.startswith('#')]
        # Only include valid lines
        lines = [line for line in lines if line.startswith('[')]

        segments = []
        for i, line in enumerate(lines):
            # get time stamp
            start_str, text = re.search(r'\[(\d+:\d+(?:\.\d+)?)](.*)', line).group(1, 2)
            start = parse_timestamp(start_str, fmt='lrc')

            if i != len(lines) - 1:
                end_str = re.search(r'\[(\d+:\d+(?:\.\d+)?)]', lines[i + 1]).group(1)
                end = parse_timestamp(end_str, fmt='lrc')
            else:
                end = None

            if not text.strip():
                continue

            segments.append({'start': start, 'end': end, 'text': text.strip()})

        lang = detect_lang(' '.join(segment['text'] for segment in segments[:10]))

        return cls(language=lang, segments=segments, filename=filename)

    @classmethod
    def from_srt(cls, filename):
        """
        Processes an SRT (SubRip Subtitle) file according to the SRT specifications outlined
        at http://www.textfiles.com/uploads/kds-srt.txt.

        The SRT format consists of:
            - A sequential subtitle number.
            - The start time and end time of the subtitle display, separated by '-->'.
            - The subtitle text, which can span one or more lines.
            - A blank line indicating the end of a subtitle entry.

        This function is designed to read or manipulate an SRT file based on the provided filename.

        Args:
            filename (str): The path to the SRT file to be processed.

        Returns:
            The return value is not specified in the provided docstring. Depending on the implementation,
            this function could return a data structure representing the parsed SRT file, a success status,
            or possibly nothing.
        """
        filename = Path(filename)
        with open(filename, encoding='utf-8') as f:
            lines = f.readlines()

        # # Remove comments
        # lines = [line for line in lines if not line.startswith('#')]

        segments = []
        i = 0
        while i < len(lines):
            line = lines[i]
            if line.strip().isdigit():
                start_str, end_str = re.search(r'(\d+:\d+:\d+,\d+) --> (\d+:\d+:\d+,\d+)', lines[i + 1]).group(1, 2)
                start = parse_timestamp(start_str, fmt='srt')
                end = parse_timestamp(end_str, fmt='srt')
                i += 2

                # Multi-line subtitle
                text = []
                while lines[i].strip():
                    text.append(lines[i].strip())
                    i += 1

                text = '\n'.join(text)
                segments.append({'start': start, 'end': end, 'text': text})

                i += 1
            else:
                raise ValueError(f'Invalid srt file {filename}')

        lang = detect_lang(' '.join(segment['text'] for segment in segments[:10]))

        return cls(language=lang, segments=segments, filename=filename)


@dataclass
class BilingualElement:
    """
    Save a LRC format element.
    """
    start: float
    end: Union[float, None]
    src_text: str
    target_text: str

    @property
    def duration(self):
        if self.end:
            return self.end - self.start
        else:
            return sys.maxsize  # Fake int infinity

    def to_json(self):
        return {'start': self.start, 'end': self.end, 'src': self.src_text, 'target': self.target_text}


class BilingualSubtitle:
    def __init__(self, src: Subtitle, target: Subtitle, filename: Union[str, Path]):
        if len(src) != len(target):
            raise ValueError(f'Source and target subtitle length not equal: {len(src)} vs {len(target)}')

        self.segments = []
        for src_seg, target_seg in zip(src.segments, target.segments):
            if src_seg.start != target_seg.start or src_seg.end != target_seg.end:
                raise ValueError(
                    f'Source and target subtitle start time not equal: {src_seg.start} vs {target_seg.start}')
            self.segments.append(BilingualElement(src_seg.start, src_seg.end, src_seg.text, target_seg.text))

        self.filename = Path(filename)
        self.suffix = '.lrc'

    @classmethod
    def from_preprocessed(cls, preprocessed_folder, audio_name):
        src_file = preprocessed_folder / f'{audio_name}_preprocessed_transcribed_optimized.json'
        target_file = preprocessed_folder / f'{audio_name}_preprocessed_transcribed_optimized_translated.json'

        if not src_file.exists() or not target_file.exists():
            raise ValueError(f'Preprocessed file not found for {audio_name}')

        src_sub = Subtitle.from_json(src_file)
        target_sub = Subtitle.from_json(target_file)

        bilingual_sub_name = preprocessed_folder / f'{audio_name}_bilingual.json'

        return cls(src_sub, target_sub, filename=bilingual_sub_name)

    def to_lrc(self):
        # If duration larger than 1 hour, use srt file instead
        if self.segments[-1].end >= 3600:
            logger.warning('Duration larger than 1 hour, use srt file instead')
            self.to_srt()
            return self.filename.with_suffix('.srt')

        lrc_path = self.filename.with_suffix('.lrc')
        fmt = partial(format_timestamp, fmt='lrc')
        with open(lrc_path, 'w', encoding='utf-8') as f:
            for i, segment in enumerate(self.segments):
                print(
                    f'[{fmt(segment.start)}] {segment.target_text.strip()}\n[{fmt(segment.start)}] {segment.src_text.strip()}',
                    file=f,
                    flush=True,
                )
                if i == len(self.segments) - 1 or segment.end != self.segments[i + 1].start:
                    print(f'[{fmt(segment.end)}]', file=f, flush=True)

        logger.info(f'File saved to {lrc_path}')

    def to_srt(self):
        srt_path = self.filename.with_suffix('.srt')
        fmt = partial(format_timestamp, fmt='srt')
        with open(srt_path, 'w', encoding='utf-8') as f:
            for i, segment in enumerate(self.segments, start=1):
                print(f'{i}\n'
                      f'{fmt(segment.start)} --> {fmt(segment.end)}\n'
                      f'{segment.target_text.strip()}\n'
                      f'{segment.src_text.strip()}\n', file=f, flush=True)

        logger.info(f'File saved to {srt_path}')
        self.suffix = '.srt'



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import re

from langcodes import Language
from lingua import LanguageDetectorBuilder

from openlrc.logger import logger

# instruction prompt modified from https://github.com/machinewrapped/gpt-subtrans
base_instruction = f'''Ignore all previous instructions.
You are a translator tasked with revising and translating subtitles into a target language. Your goal is to ensure accurate, concise, and natural-sounding translations for each line of dialogue. The input consists of transcribed audio, which may contain transcription errors. Your task is to first correct any errors you find in the sentences based on their context, and then translate them to the target language according to the revised sentences.
The user will provide a chunk of lines, you should respond with an accurate, concise, and natural-sounding translation for the dialogue. 
The user may provide additional context, such as background, description or title of the source material, a summary of the current scene, or a list of character names. Use this information to improve the quality of your translation.
Your response will be processed by an automated system, so it is imperative that you adhere to the required output format.

Example input (Japanese to Chinese):

#200
Original>
変わりゆく時代において、
Translation>

#501
Original>
生き残る秘訣は、進化し続けることです。
Translation>

You should respond with:

#200
Original>
変わく時代いて、
Translation>
在变化的时代中，

#501
Original>
生き残る秘訣は、進化し続けることです。
Translation>
生存的秘诀是不断进化。

Example input (English to German):

#700
Original>
those who resist change may find themselves left behind.
Translation>

#701
Original>
those resist change find themselves left.
Translation>

You should respond with:

#700
Original>
In the age of digital transformation,
Translation>
Im Zeitalter der digitalen Transformation,

#701
Original>
those who resist change may find themselves left behind.
Translation>
diejenigen, die sich dem Wandel widersetzen, könnten sich zurückgelassen finden.

Please ensure that each line of dialogue remains distinct in the  translation. Merging lines together can lead to timing problems during playback.

At the end of each set of translations, include a one or two line synopsis of the input text in a <summary/> tag, for example:
<summary>John and Sarah discuss their plan to locate a suspect, deducing that he is likely in the uptown area.</summary>
Remember to end this tag with ``</summary>``.

Use the available information to add a short description of the current scene in a <scene/> tag, for example:
<scene>John and Sarah are in their office analyzing data and planning their next steps. They deduce that the suspect is probably in the uptown area and decide to start their search there.</scene>
Remember to end this tag with ``</scene>``.

Use the target language when writing content for the <summary/> and <scene/> tags. 
Ensure that the summary and scene are concise, containing less than 100 words.
You need to update your summary and scene with the new information you have.
Do not guess or improvise if the context is unclear, just summarise the dialogue.

The translation should be in a lovely colloquial style and suitable for high-quality subtitles.

I’m going to tip $1000 for a better translation!

#######################
There was an issue with the previous translation. 

Remember to include ``<summary>`` and ``<scene>`` tags in your response.
Do not translate ``Original>`` and ``Translation>``.
Please translate the subtitles again, paying careful attention to ensure that each line is translated separately, and that every line has a matching translation.
Do not merge lines together in the translation, it leads to incorrect timings and confusion for the reader.
The content of the translation is for learning purposes only and will not violate the usage guidelines. '''


class TranslatePrompter:
    @classmethod
    def format_texts(cls, texts):
        raise NotImplementedError()

    @staticmethod
    def post_process(texts):
        raise NotImplementedError()

    def check_format(self, messages, output_str):
        raise NotImplementedError()


class BaseTranslatePrompter(TranslatePrompter):
    def __init__(self, src_lang, target_lang, audio_type=None, title='', background='', description=''):
        self.src_lang = src_lang
        self.target_lang = target_lang
        self.src_lang_display = Language.get(src_lang).display_name('en')
        self.target_lang_display = Language.get(target_lang).display_name('en')
        self.lan_detector = LanguageDetectorBuilder.from_all_languages().build()

        self.audio_type = audio_type
        self.title = title
        self.background = background
        self.description = description
        self.user_prompt = f'''
{f"<title>{self.title}</title>" if self.title else ""}
{f"<background>{self.background}</background>" if self.background else ""}
{f"<description>{self.description}</description>" if self.description else ""}
<context>
<scene>{{scene}}</scene>
<chunk> {{summaries_str}} </chunk>
</context>
<chunk_id> Scene 1 Chunk {{chunk_num}} <chunk_id>

Please translate these subtitles for {self.audio_type}{f" named {self.title}" if self.title else ""} from {self.src_lang_display} to {self.target_lang_display}.\n
{{user_input}}

<summary></summary>
<scene></scene>'''

    @staticmethod
    def system():
        return base_instruction

    def user(self, chunk_num, user_input, summaries='', scene=''):
        summaries_str = '\n'.join(f'Chunk {i}: {summary}' for i, summary in enumerate(summaries, 1))
        return self.user_prompt.format(summaries_str=summaries_str, scene=scene,
                                       chunk_num=chunk_num, user_input=user_input).strip()

    @classmethod
    def format_texts(cls, texts):
        """
        Reconstruct list of text into desired format.

        Args:
            texts: List of (id, text).

        Returns:
            The formatted string: f"#id\nOriginal>\n{text}\nTranslation>\n"
        """
        return '\n'.join([f'#{i}\nOriginal>\n{text}\nTranslation>\n' for i, text in texts])

    def check_format(self, messages, content):
        summary = re.search(r'<summary>(.*)</summary>', content)
        scene = re.search(r'<scene>(.*)</scene>', content)

        # If message is for claude, use messages[0]
        user_input = messages[1]['content'] if len(messages) == 2 else messages[0]['content']

        original = re.findall(r'Original>\n(.*?)\nTranslation>', user_input, re.DOTALL)
        translation = re.findall(r'Translation>\n*(.*?)(?:#\d+|<summary>|\n*$)', content, re.DOTALL)

        if not original or not translation:
            # TODO: Try to change chatbot_model if always fail

            logger.warning(f'Fail to extract original or translation.')
            logger.debug(f'Content: {content}')
            return False

        if len(original) != len(translation):
            logger.warning(
                f'Fail to ensure length consistent: original is {len(original)}, translation is {len(translation)}')
            logger.debug(f'original: {original}')
            logger.debug(f'translation: {original}')
            return False

        # Ensure the translated langauge is in the target language
        if len(translation) >= 3:
            # 3-voting for detection stability
            chunk_size = len(translation) // 3
            translation_chunks = [translation[i:i + chunk_size] for i in range(0, len(translation), chunk_size)]
            if len(translation_chunks) > 3:
                translation_chunks[-2].extend(translation_chunks[-1])
                translation_chunks.pop()

            translated_langs = [self.lan_detector.detect_language_of(' '.join(chunk)) for chunk in translation_chunks]
            translated_langs = [lang.name.lower() for lang in translated_langs if lang]

            if not translated_langs:
                # Cant detect language
                return True

            # get the most common language
            translated_lang = max(set(translated_langs), key=translated_langs.count)
        else:
            detected_lang = self.lan_detector.detect_language_of(' '.join(translation))
            if not detected_lang:
                # Cant detect language
                return True
            translated_lang = detected_lang.name.lower()

        target_lang = Language.get(self.target_lang).language_name().lower()
        if translated_lang != target_lang:
            logger.warning(f'Translated language is {translated_lang}, not {target_lang}.')
            return False

        # It's ok to keep going without summary and scene
        if not summary or not summary.group(1):
            logger.warning(f'Fail to extract summary.')
        if not scene or not scene.group(1):
            logger.warning(f'Fail to extract scene.')

        return True


class AtomicTranslatePrompter(TranslatePrompter):
    def __init__(self, src_lang, target_lang):
        self.src_lang = src_lang
        self.target_lang = target_lang
        self.src_lang_display = Language.get(src_lang).display_name('en')
        self.target_lang_display = Language.get(target_lang).display_name('en')
        self.lan_detector = LanguageDetectorBuilder.from_all_languages().build()

    def user(self, text):
        return f'''Please translate the following text from {self.src_lang_display} to {self.target_lang_display}. 
Please do not output any content other than the translated text. Here is the text: {text}'''

    def check_format(self, messages, output_str):
        # Ensure the translated langauge is in the target language
        detected_lang = self.lan_detector.detect_language_of(output_str)
        if not detected_lang:
            # Cant detect language
            return True

        translated_lang = detected_lang.name.lower()
        target_lang = Language.get(self.target_lang).language_name().lower()
        if translated_lang != target_lang:
            logger.warning(f'Translated text: "{output_str}" is {translated_lang}, not {target_lang}.')
            return False

        return True


prompter_map = {
    'base_trans': BaseTranslatePrompter,
}



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.
import logging
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import List, Union

import torch
from df.enhance import enhance, init_df, load_audio, save_audio
from ffmpeg_normalize import FFmpegNormalize
from tqdm import tqdm

from openlrc.defaults import default_preprocess_options
from openlrc.logger import logger
from openlrc.utils import release_memory


def loudness_norm_single(audio_path: Path, ln_path: Path):
    """
    Normalize the loudness of a single audio file using FFmpegNormalize.

    Args:
        audio_path (Path): The path to the input audio file.
        ln_path (Path): The path to save the normalized audio file.
    """
    normalizer = FFmpegNormalize(output_format='wav', sample_rate=48000, progress=logger.level <= logging.DEBUG,
                                 keep_lra_above_loudness_range_target=True)

    if not ln_path.exists():
        normalizer.add_media_file(str(audio_path), str(ln_path))
        normalizer.run_normalization()


class Preprocessor:
    """
    Preprocess audio to make it clear and normalized.
    """

    def __init__(self, audio_paths: Union[str, Path, List[str], List[Path]], output_folder='preprocessed',
                 options: dict = default_preprocess_options):
        if not isinstance(audio_paths, list):
            audio_paths = [audio_paths]
        self.audio_paths = [Path(p) for p in audio_paths]
        self.output_paths = [p.parent / output_folder for p in self.audio_paths]
        self.options = options

        for path in self.output_paths:
            if not path.exists():
                path.mkdir()

    def noise_suppression(self, audio_paths: Union[str, Path, List[str], List[Path]], atten_lim_db=15):
        """
        Supress noise in audio.
        """
        if not audio_paths:
            return []

        if 'atten_lim_db' in self.options.keys():
            atten_lim_db = self.options['atten_lim_db']

        model, df_state, _ = init_df()

        ns_audio_paths = []
        for audio_path, output_path in zip(audio_paths, self.output_paths):
            audio_name = audio_path.stem
            ns_path = output_path / f'{audio_name}_ns.wav'

            if not ns_path.exists():
                audio, info = load_audio(audio_path, sr=df_state.sr())

                # Split audio into 10 min chunks
                audio_chunks = [audio[:, i:i + 600 * info.sample_rate]
                                for i in range(0, audio.shape[1], 600 * info.sample_rate)]

                enhanced_chunks = []
                for ac in tqdm(audio_chunks, desc=f'Noise suppressing for {audio_name}'):
                    enhanced_chunks.append(enhance(model, df_state, ac, atten_lim_db=atten_lim_db))

                enhanced = torch.cat(enhanced_chunks, dim=1)

                assert enhanced.shape == audio.shape, 'Enhanced audio shape does not match original audio shape.'

                save_audio(ns_path, enhanced, sr=df_state.sr())

            ns_audio_paths.append(ns_path)

        release_memory(model)

        return ns_audio_paths

    def loudness_normalization(self, audio_paths: Union[str, Path, List[str], List[Path]]):
        """
        Normalize loudness of audio.
        """
        logger.info('Loudness normalizing...')

        args = []
        ln_audio_paths = []
        for audio_path, output_path in zip(audio_paths, self.output_paths):
            ln_path = output_path / f'{audio_path.stem}_ln.wav'
            args.append((audio_path, ln_path))
            ln_audio_paths.append(ln_path)

        # Multi-processing
        with ProcessPoolExecutor() as executor:
            results = [executor.submit(loudness_norm_single, *arg) for arg in args]

            exceptions = [res.exception() for res in results]
            if any(exceptions):
                # Get the first not None exception
                exception = next(filter(None, exceptions))

                logger.error(f'Loudness normalization failed, exception: {exception}')
                raise exception

        return ln_audio_paths

    def run(self, noise_suppress=False):
        """
        Args:
            noise_suppress (bool, optional): A boolean flag indicating whether to perform noise suppression.
                Defaults to False.

        Returns:
            list of Path: A list of Path objects representing the final processed audio paths.
        """
        # Check if the preprocessed audio already exists.
        need_process = []
        final_processed = []
        for audio_path, output_path in zip(self.audio_paths, self.output_paths):
            audio_name = audio_path.stem
            preprocessed_path = output_path / f'{audio_name}_preprocessed.wav'
            final_processed.append(preprocessed_path)
            if preprocessed_path.exists():
                logger.info(f'Preprocessed audio already exists in {preprocessed_path}')
                continue
            else:
                need_process.append(audio_path)

        ns_paths = need_process
        if noise_suppress:
            ns_paths = self.noise_suppression(need_process)
        ln_paths: list[Path] = self.loudness_normalization(ns_paths)

        for ln_path, audio_path in zip(ln_paths, need_process):
            audio_name = audio_path.stem

            ln_path = ln_path.rename(ln_path.parent / f'{audio_name}_preprocessed.wav')
            logger.info(f'Preprocessed audio saved to {ln_path}')

        return final_processed



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import subprocess
import sys


def run_gui():
    try:
        # Assuming the Streamlit app is located at "openlrc/gui/home.py"
        subprocess.run(["streamlit", "run", "openlrc/gui/home.py"], check=True)
    except subprocess.CalledProcessError as e:
        print(f"Failed to launch the Streamlit app: {e}", file=sys.stderr)
        sys.exit(1)


def main():
    if len(sys.argv) > 1 and sys.argv[1] == "gui":
        run_gui()
    else:
        print("Usage: openlrc gui", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import concurrent.futures
import json
import shutil
from pathlib import Path
from pprint import pformat
from queue import Queue
from threading import Lock
from typing import List

from faster_whisper.transcribe import Segment

from openlrc.context import Context
from openlrc.defaults import default_asr_options, default_vad_options, default_preprocess_options
from openlrc.logger import logger
from openlrc.opt import SubtitleOptimizer
from openlrc.preprocess import Preprocessor
from openlrc.subtitle import Subtitle, BilingualSubtitle
from openlrc.transcribe import Transcriber
from openlrc.translate import LLMTranslator
from openlrc.utils import Timer, extend_filename, get_audio_duration, format_timestamp, extract_audio, \
    get_file_type


class LRCer:
    """
    Args:
        whisper_model: Name of whisper model (tiny, tiny.en, base, base.en, small, small.en, medium,
            medium.en, large-v1, large-v2, large-v3, distill-large-v3) When a size is configured,
            the converted model is downloaded from the Hugging Face Hub. Default: ``large-v3``
        compute_type: The type of computation to use. Can be ``int8``, ``int8_float16``, ``int16``,
            ``float16`` or ``float32``. Default: ``float16``
        chatbot_model: The chatbot model to use, currently we support gptbot from , claudebot from Anthropic.
            OpenAI: gpt-4-0125-preview, gpt-4-turbo-preview, gpt-3.5-turbo-0125, gpt-3.5-turbo
            Anthropic: claude-3-opus-20240229, claude-3-sonnet-20240229, claude-3-haiku-20240307
            Default: ``gpt-3.5-turbo``
        fee_limit: The maximum fee you are willing to pay for one translation call. Default: ``0.1``
        consumer_thread: To prevent exceeding the RPM and TPM limits set by OpenAI, the default is TPM/MAX_TOKEN.
        asr_options: Parameters for whisper model.
        vad_options: Parameters for VAD model.
        proxy: Proxy for openai requests. e.g. 'http://127.0.0.1:7890'
        base_url_config: Base URL dict for OpenAI & Anthropic.
            e.g. {'openai': 'https://openai.justsong.cn/', 'anthropic': 'https://api.g4f.icu'}
            Default: ``None``
    """

    def __init__(self, whisper_model='large-v3', compute_type='float16', chatbot_model: str = 'gpt-3.5-turbo',
                 fee_limit=0.1, consumer_thread=4, asr_options=None, vad_options=None, preprocess_options=None,
                 proxy=None, base_url_config=None):
        self.chatbot_model = chatbot_model
        self.fee_limit = fee_limit
        self.api_fee = 0  # Can be updated in different thread, operation should be thread-safe
        self.from_video = set()
        self.context: Context = Context()
        self.proxy = proxy
        self.base_url_config = base_url_config

        self._lock = Lock()
        self.exception = None
        self.consumer_thread = consumer_thread

        # Default Automatic Speech Recognition (ASR) options
        self.asr_options = default_asr_options

        # Parameters for VAD (see faster_whisper.vad.VadOptions), tune them if speech is not being detected
        self.vad_options = default_vad_options

        self.preprocess_options = default_preprocess_options

        if asr_options:
            self.asr_options.update(asr_options)

        if vad_options:
            self.vad_options.update(vad_options)

        if preprocess_options:
            self.preprocess_options.update(preprocess_options)

        self.transcriber = Transcriber(model_name=whisper_model, compute_type=compute_type,
                                       asr_options=self.asr_options, vad_options=self.vad_options)
        self.transcribed_paths = []

    def transcription_producer(self, transcription_queue, audio_paths, src_lang):
        """
        Sequential Producer.
        """
        for audio_path in audio_paths:
            transcribed_path = extend_filename(audio_path, '_transcribed').with_suffix('.json')
            if not transcribed_path.exists():
                with Timer('Transcription process'):
                    logger.info(
                        f'Audio length: {audio_path}: {format_timestamp(get_audio_duration(audio_path), fmt="srt")}')
                    segments, info = self.transcriber.transcribe(audio_path, language=src_lang)
                    logger.info(f'Detected language: {info.language}')

                    # [Segment(start, end, text, words=[Word(start, end, word, probability)])]

                # Save the transcribed json
                self.to_json(segments, name=transcribed_path, lang=info.language)  # xxx_transcribed.json
            else:
                logger.info(f'Found transcribed json file: {transcribed_path}')
            transcription_queue.put(transcribed_path)
            # logger.info(f'Put transcription: {transcribed_path}')

        transcription_queue.put(None)
        logger.info('Transcription producer finished.')

    def transcription_consumer(self, transcription_queue, target_lang, prompter, skip_trans, bilingual_sub):
        """
        Parallel Consumer.
        """
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [executor.submit(self.consumer_worker, transcription_queue, target_lang, prompter, skip_trans,
                                       bilingual_sub)
                       for _ in range(self.consumer_thread)]
            concurrent.futures.wait(futures)
        logger.info('Transcription consumer finished.')

    def consumer_worker(self, transcription_queue, target_lang, prompter, skip_trans, bilingual_sub):
        """
        Parallel translation.
        """
        while True:
            logger.debug(f'Translation worker waiting transcription...')
            transcribed_path = transcription_queue.get()

            if transcribed_path is None:
                transcription_queue.put(None)
                logger.debug('Translation worker finished.')
                return

            logger.info(f'Got transcription: {transcribed_path}')
            transcribed_sub = Subtitle.from_json(transcribed_path)
            transcribed_opt_sub = self.post_process(transcribed_sub, update_name=True)
            audio_name = transcribed_path.stem.replace('_transcribed', '')

            # xxx_transcribed_optimized_translated.json
            translated_path = extend_filename(transcribed_opt_sub.filename, '_translated')

            final_json_path = Path(translated_path.parent / f'{audio_name}.json')
            if final_json_path.exists():
                final_subtitle = Subtitle.from_json(final_json_path)
            elif skip_trans:
                shutil.copy(transcribed_opt_sub.filename, final_json_path)
                final_subtitle = transcribed_opt_sub
                final_subtitle.filename = final_json_path
            else:
                with Timer('Translation process'):
                    try:
                        final_subtitle = self._translate(audio_name, prompter, target_lang, transcribed_opt_sub,
                                                         translated_path)
                    except Exception as e:
                        self.exception = e
                        return

            # Copy preprocessed/xxx_preprocessed.lrc or preprocessed/xxx_preprocessed.srt to xxx.lrc or xxx.srt
            if audio_name in self.from_video:
                subtitle_path = final_subtitle.to_srt()
            else:
                subtitle_path = final_subtitle.to_lrc()

            suffix = subtitle_path.suffix
            result_path = subtitle_path.parents[1] / subtitle_path.name.replace(f'_preprocessed{suffix}',
                                                                                suffix)
            shutil.copy(subtitle_path, result_path)

            # Bilingual
            if bilingual_sub:
                bilingual_subtitle = BilingualSubtitle.from_preprocessed(transcribed_path.parent,
                                                                         audio_name.replace('_preprocessed', ''))
                bilingual_subtitle.to_lrc()
                bilingual_lrc_path = bilingual_subtitle.filename.with_suffix(bilingual_subtitle.suffix)
                shutil.copy(bilingual_lrc_path, result_path.parent / bilingual_lrc_path.name)

            logger.info(f'Translation fee til now: {self.api_fee:.4f} USD')

            self.transcribed_paths.append(result_path)

    def _translate(self, audio_name, prompter, target_lang, transcribed_opt_sub, translated_path):
        json_filename = Path(translated_path.parent / (audio_name + '.json'))
        compare_path = Path(translated_path.parent, f'{audio_name}_compare.json')
        if not translated_path.exists():
            # Translate the transcribed json
            translator = LLMTranslator(chatbot_model=self.chatbot_model, prompter=prompter, fee_limit=self.fee_limit,
                                       proxy=self.proxy, base_url_config=self.base_url_config)
            context = self.context

            target_texts = translator.translate(
                transcribed_opt_sub.texts,
                src_lang=transcribed_opt_sub.lang,
                target_lang=target_lang,
                title=audio_name,
                audio_type=context.audio_type,
                background=context.background,
                description=context.get_description(audio_name),
                compare_path=compare_path
            )

            with self._lock:
                self.api_fee += translator.api_fee  # Ensure thread-safe

            transcribed_opt_sub.set_texts(target_texts, lang=target_lang)

            # xxx_transcribed_optimized_translated.json
            transcribed_opt_sub.save(translated_path, update_name=True)
        else:
            logger.info(f'Found translated json file: {translated_path}')
        translated_sub = Subtitle.from_json(translated_path)

        final_subtitle = self.post_process(translated_sub, output_name=json_filename, update_name=True)  # xxx.json

        return final_subtitle

    def run(self, paths, src_lang=None, target_lang='zh-cn', prompter='base_trans', context_path=None,
            skip_trans=False, noise_suppress=False, bilingual_sub=False, clear_temp_folder=False) -> List[str]:
        """
        Split the translation into 2 phases: transcription and translation. They're running in parallel.
        Firstly, transcribe the audios one-by-one. At the same time, translation threads are created and waiting for
        the transcription results. After all the transcriptions are done, the translation threads will start to
        translate the transcribed texts.

        Args:
            paths (Union[str, Path, List[Union[str, Path]]]): Audio/Video paths, can be a list or a single path.
            src_lang (str): Language of the audio, default to auto-detect.
            target_lang (str): Target language, default to Mandarin Chinese.
            prompter (str): Currently, only `base_trans` is supported.
            context_path (str): path to context config file. (Default to use `context.yaml` in the first audio's directory)
            skip_trans (bool): Whether to skip the translation process. (Default to False)
            noise_suppress (bool): Whether to suppress the noise in the audio. (Default to False)
            bilingual_sub (bool): Whether to generate bilingual subtitles. (Default to False)
            clear_temp_folder (bool): Whether to clear the temporary folder.
                Note, set this back to False to see more intermediate results if error encountered. (Default to False)

        Returns:
            List[str]: List of paths to the transcribed files.

        Raises:
            Exception: If an exception occurs during the transcription or translation process.
        """
        self.transcribed_paths = []

        if not paths:
            logger.warning('No audio/video file given. Skip LRCer.run()')
            return []

        if isinstance(paths, str) or isinstance(paths, Path):
            paths = [paths]

        paths = list(map(Path, paths))

        if context_path:
            context_path = Path(context_path)
            self.context.load_config(context_path)
            logger.info(f'Found context config: {context_path}')
            logger.debug(f'Context: {self.context}')
        else:
            # Try to find the default `context.yaml` in the first audio's directory
            try:
                context_path = paths[0].parent / 'context.yaml'
                self.context.load_config(context_path)
                logger.info(f'Found context config file: {context_path}')
            except FileNotFoundError:
                logger.info(f'Default context config not found: {self.context}, using default context.')

        audio_paths = self.pre_process(paths, noise_suppress=noise_suppress)

        logger.info(f'Working on {len(audio_paths)} audio files: {pformat(audio_paths)}')

        transcription_queue = Queue()

        with Timer('Transcription (Producer) and Translation (Consumer) process'):
            consumer = concurrent.futures.ThreadPoolExecutor(thread_name_prefix='Consumer') \
                .submit(self.transcription_consumer, transcription_queue, target_lang, prompter, skip_trans,
                        bilingual_sub)
            producer = concurrent.futures.ThreadPoolExecutor(thread_name_prefix='Producer') \
                .submit(self.transcription_producer, transcription_queue, audio_paths, src_lang)

            producer.result()
            consumer.result()

            if self.exception:
                # traceback.print_exception(type(self.exception), self.exception, self.exception.__traceback__)
                raise self.exception

        logger.info(f'Totally used API fee: {self.api_fee:.4f} USD')

        if clear_temp_folder:
            logger.info('Clearing temporary folder...')
            self.clear_temp_files(audio_paths)

        return self.transcribed_paths

    @staticmethod
    def clear_temp_files(paths):
        """
        Clear the temporary files generated during the transcription and translation process.
        """
        temp_folders = set([path.parent for path in paths])
        for folder in temp_folders:
            assert folder.name == 'preprocessed', f'Not a temporary folder: {folder}'

            shutil.rmtree(folder)
            logger.debug(f'Removed {folder}')

    @staticmethod
    def to_json(segments: List[Segment], name, lang):
        result = {
            'language': lang,
            'segments': []
        }

        for segment in segments:
            result['segments'].append({
                'start': segment.start,
                'end': segment.end,
                'text': segment.text
            })

        with open(name, 'w', encoding='utf-8') as f:
            json.dump(result, f, ensure_ascii=False, indent=4)

        logger.info(f'File saved to {name}')

        return result

    def pre_process(self, paths, noise_suppress=False):
        paths = [Path(path) for path in paths]

        # Check if path is audio or video
        for i, path in enumerate(paths):
            if not path.exists() or not path.is_file():
                raise FileNotFoundError(f'File not found: {path}')

            paths[i] = extract_audio(path)

            if get_file_type(path) == 'video':
                self.from_video.add(path.stem + '_preprocessed')

        # Audio-based process
        preprocessor = Preprocessor(paths, options=self.preprocess_options)
        paths = preprocessor.run(noise_suppress)

        return paths

    @staticmethod
    def post_process(transcribed_sub: Path, output_name: Path = None, remove_files: List[Path] = None,
                     update_name=False):
        optimizer = SubtitleOptimizer(transcribed_sub)
        optimizer.perform_all()
        optimizer.save(output_name, update_name=update_name)

        # Remove intermediate files
        if remove_files:
            _ = [file.unlink() for file in remove_files if file.is_file()]

        return optimizer.subtitle



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import re
from pathlib import Path
from typing import Union

import zhconv

from openlrc.logger import logger
from openlrc.subtitle import Subtitle
from openlrc.utils import extend_filename

# Different language may need different threshold
cut_long_threshold = {
    350: 'en',
    125: 'cn, ja'
}


class SubtitleOptimizer:
    """
    SubtitleOptimizer class is used to optimize subtitles by performing various operations. For example, merging same
    text, merging short duration subtitles, merging repeated patterns, cutting long texts, converting traditional
    Chinese to Mandarin, optimizing punctuation, removing '<unk>' tags, removing empty elements, and stripping
    whitespace.

    Args:
        subtitle (Union[Path, Subtitle]): The subtitle to be optimized.
    """

    def __init__(self, subtitle: Union[Path, Subtitle]):
        if isinstance(subtitle, Path):
            subtitle = Subtitle.from_json(subtitle)

        self.subtitle = subtitle
        self.lang = self.subtitle.lang

    @property
    def filename(self):
        return self.subtitle.filename

    def merge_same(self):
        """
        Merge the same text.
        """
        new_elements = []

        for i, element in enumerate(self.subtitle.segments):
            if i == 0 or element.text != new_elements[-1].text:
                new_elements.append(element)
            else:
                new_elements[-1].end = element.end

        self.subtitle.segments = new_elements

    def merge_short(self, threshold=1.2):
        """
        Merge the short duration subtitle.
        """
        new_elements = []

        merged_element = None
        for i, element, in enumerate(self.subtitle.segments):
            if i == 0 or element.duration >= threshold:
                if merged_element:
                    if merged_element.duration < 1.5:
                        # If the merged elements is still too small, find the closer element nearby and merge it
                        if element.start - merged_element.end < merged_element.start - new_elements[-1].end:
                            # merge to the next
                            element.text = merged_element.text + element.text
                            element.start = merged_element.start
                        else:
                            # merge to the last
                            new_elements[-1].text += merged_element.text
                            new_elements[-1].end = merged_element.end
                    else:
                        new_elements.append(merged_element)
                    new_elements.append(element)
                    merged_element = None
                else:
                    new_elements.append(element)
            else:
                if not merged_element:
                    merged_element = element
                else:
                    merged_element.text += element.text
                    merged_element.end = element.end

        self.subtitle.segments = new_elements

    def merge_repeat(self):
        """
        Merge the same pattern in one lyric.
        :return:
        """
        new_elements = self.subtitle.segments

        for i, element in enumerate(new_elements):
            text = element.text
            text = re.sub(r'(.)\1{4,}', r'\1\1...', text)
            text = re.sub(r'(.+)\1{4,}', r'\1\1...', text)
            new_elements[i].text = text

        self.subtitle.segments = new_elements

    def cut_long(self, keep=20):
        threshold = 150
        for threshold, lang in cut_long_threshold.items():
            if self.lang.lower() in lang:
                threshold = threshold
                break

        new_elements = self.subtitle.segments

        for i, element in enumerate(new_elements):
            if len(element.text) > threshold and len(element.text) / len(set(element.text)) > 3.0:
                # text length larger than threshold and text density is larger than 3.0
                logger.warning(f'Cut long text: {element.text}\nInto: {element.text[:keep]}...')
                new_elements[i].text = element.text[:keep]

        self.subtitle.segments = new_elements

    def traditional2mandarin(self):
        new_elements = self.subtitle.segments

        for i, element in enumerate(new_elements):
            new_elements[i].text = zhconv.convert(element.text, locale='zh-cn')

        self.subtitle.segments = new_elements

    def punctuation_optimization(self):
        def replace_punctuation_with_chinese(text):
            # Mapping of English punctuation to Chinese punctuation
            punctuation_mapping = {
                ',': '，',
                '.': '。',
                '?': '？',
                '!': '！',
                ':': '：',
                ';': '；',
                '"': '”',
                "'": '’',
                '(': '（',
                ')': '）',
                '[': '【',
                ']': '】',
                '{': '｛',
                '}': '｝'
            }

            # Compile a regular expression pattern for all English punctuation marks
            pattern = re.compile("|".join(map(re.escape, punctuation_mapping.keys())))

            # Replace the punctuation in the text
            result = pattern.sub(lambda match: punctuation_mapping[match.group(0)], text)

            # Replace consecutive (>3) 。 with ...
            result = re.sub(r'(。){3,}', '...', result)

            return result

        new_elements = self.subtitle.segments

        for i, element in enumerate(new_elements):
            new_elements[i].text = replace_punctuation_with_chinese(element.text)

        self.subtitle.segments = new_elements

    def remove_unk(self):
        new_elements = self.subtitle.segments

        for i, element in enumerate(new_elements):
            new_elements[i].text = element.text.replace('<unk>', '')

        self.subtitle.segments = new_elements

    def remove_empty(self):
        self.subtitle.segments = [element for element in self.subtitle.segments if element.text]

    def strip(self):
        new_elements = self.subtitle.segments

        for i, element in enumerate(new_elements):
            new_elements[i].text = element.text.strip()

        self.subtitle.segments = new_elements

    def perform_all(self):
        for _ in range(2):
            self.merge_same()
            self.merge_short()
            self.merge_repeat()
            self.cut_long()
            self.remove_unk()
            self.remove_empty()
            self.strip()

            if self.subtitle.lang.lower() in ['zh-cn', 'zh']:
                self.traditional2mandarin()
            if self.subtitle.lang.lower() in ['zh-cn', 'zh', 'zh-tw']:
                self.punctuation_optimization()

    def save(self, output_name=None, update_name=False):
        optimized_name = extend_filename(self.filename, '_optimized') if not output_name else output_name
        self.subtitle.save(optimized_name, update_name=update_name)
        logger.info(f'Optimized json file saved to {optimized_name}')



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import re
import subprocess
import time
from pathlib import Path
from typing import List, Dict, Any, Union

import audioread
import ffmpeg
import filetype
import jaconvV2
import langcodes
import spacy
import tiktoken
import torch
import unicodedata
from lingua import LanguageDetectorBuilder

from openlrc.defaults import supported_languages_lingua
from openlrc.logger import logger


def extract_audio(path: Path) -> Path:
    """
    Extract audio from video.
    :return: Audio path
    """
    file_type = get_file_type(path)
    if file_type == 'audio':
        return path

    probe = ffmpeg.probe(path)
    audio_streams = next((stream for stream in probe['streams'] if stream['codec_type'] == 'audio'), None)
    sample_rate = audio_streams['sample_rate']
    logger.info(f'File {path}: Audio sample rate: {sample_rate}')

    audio, err = (
        ffmpeg.input(path).
        output("pipe:", format='wav', acodec='pcm_s16le', ar=sample_rate, loglevel='quiet').
        run(capture_stdout=True)
    )

    if err:
        raise RuntimeError(f'ffmpeg error: {err}')

    audio_path = path.with_suffix('.wav')
    with open(audio_path, 'wb') as f:
        f.write(audio)

    return audio_path


def get_file_type(path: Path) -> str:
    if path.suffix == '.ts':
        return 'video'

    try:
        file_type = filetype.guess(path).mime.split('/')[0]
    except (TypeError, AttributeError) as e:
        raise RuntimeError(f'File {path} is not a valid file.') from e

    if file_type not in ['audio', 'video']:
        raise RuntimeError(f'File {path} is not a valid file. Should be audio or video file.')

    return file_type


def get_audio_duration(path: Union[str, Path]) -> float:
    with audioread.audio_open(str(path)) as audio:
        return audio.duration


def release_memory(model: torch.nn.Module) -> None:
    # gc.collect()
    torch.cuda.empty_cache()
    del model


def normalize(text):
    """
    Normalize strings using str.lower(), and unicodedata.normalize
    """
    # unicodedata cant handel quotes as expected'’'
    quotes_table = str.maketrans('〈〉゛“”‘’', '<>"""\'\'')
    text = text.translate(quotes_table)

    text = unicodedata.normalize('NFKC', text.lower())

    # unicodedata cant handel kana
    text = jaconvV2.z2h(text)

    # Special case
    special_table = str.maketrans('〇①②③④⑤⑥⑦⑧⑨', '0123456789')
    text = text.translate(special_table)

    return text


def get_text_token_number(text: str, model: str = "gpt-3.5-turbo") -> int:
    tokens = tiktoken.encoding_for_model(model).encode(text)

    return len(tokens)


def get_messages_token_number(messages: List[Dict[str, Any]], model: str = "gpt-3.5-turbo") -> int:
    total = sum([get_text_token_number(element['content'], model=model) for element in messages])

    return total


def extend_filename(filename: Path, extend: str) -> Path:
    """Extend a filename with some string."""
    return filename.with_stem(filename.stem + extend)


class Timer:
    def __init__(self, task=""):
        self._start = None
        self._stop = None
        self.task = task

    def start(self):
        if self.task:
            logger.info(f'Start {self.task}')
        self._start = time.perf_counter()

    def stop(self):
        self._stop = time.perf_counter()
        logger.info(f'{self.task} Elapsed: {self._elapsed:.2f}s')

    @property
    def _elapsed(self):
        return self._stop - self._start

    @property
    def duration(self):
        return time.perf_counter() - self._start

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()


def parse_timestamp(time_stamp: str, fmt: str) -> float:
    """
    Parse a timestamp from a subtitle file and convert it to seconds.

    Args:
        time_stamp (str): The timestamp to parse.
        fmt (str): The format of `time_stamp`. Supported values are:
            - 'lrc' for LRC format, e.g., '1:23.45'
            - 'srt' for SRT format, e.g., '01:23:45,678'

    Returns:
        float: The timestamp in seconds.

    Raises:
        ValueError: If `time_stamp` does not match the expected format for the specified `fmt`.
    """

    if fmt == 'lrc':
        if not re.match(r'^\d+:\d+\.\d+$', time_stamp):
            raise ValueError(f"Invalid timestamp format for LRC: {time_stamp}")
        minutes, seconds = time_stamp.split(':')
        seconds, hundredths_of_sec = seconds.split('.')
        return int(minutes) * 60 + int(seconds) + int(hundredths_of_sec) / 100.0
    elif fmt == 'srt':
        if not re.match(r'^\d+:\d+:\d+,\d+$', time_stamp):
            raise ValueError(f"Invalid timestamp format for SRT: {time_stamp}")
        hours, minutes, seconds = time_stamp.split(':')
        seconds, milliseconds = seconds.split(',')
        return int(hours) * 3600 + int(minutes) * 60 + int(seconds) + int(milliseconds) / 1000.0
    else:
        raise ValueError(f"Unsupported timestamp format: {fmt}")


def format_timestamp(seconds: float, fmt: str = 'lrc') -> str:
    """
    Converts a timestamp in seconds into a string in the specified format.

    Args:
        seconds (float): Timestamp in seconds.
        fmt (str): Format of the output string. Supported values are:
            - 'lrc' for LRC format, e.g., '1:23.45'
            - 'srt' for SRT format, e.g., '01:23:45,678'

    Returns:
        str: A string representation of the timestamp in the specified format.
    """
    # Ensure that the timestamp is non-negative.
    assert seconds >= 0, "non-negative timestamp expected"

    # Convert seconds into milliseconds.
    milliseconds = round(seconds * 1000.0)

    # Extract hours, minutes, seconds, and milliseconds from milliseconds.
    hours = milliseconds // 3600000
    milliseconds %= 3600000
    minutes = milliseconds // 60000
    milliseconds %= 60000
    seconds = milliseconds // 1000
    milliseconds %= 1000

    # Return the timestamp in the specified format.
    if fmt == 'lrc':
        return f"{minutes:02d}:{seconds:02d}.{milliseconds // 10:02d}"
    elif fmt == 'srt':
        return f"{hours:02d}:{minutes:02d}:{seconds:02d},{milliseconds:03d}"
    else:
        raise ValueError(f"Unsupported timestamp format: {fmt}")


def detect_lang(text):
    detector = LanguageDetectorBuilder.from_languages(*supported_languages_lingua).build()
    name = detector.detect_language_of(' '.join(text)).name.lower()
    lang_code = langcodes.Language.find(name).language
    return lang_code


def get_spacy_lib(lang):
    special_case = {
        'core_web': ['zh', 'en'],
        'ent_wiki': ['xx']
    }

    mid_str = 'core_news'
    for k, v in special_case.items():
        if lang in v:
            mid_str = k

    return f'{lang}_{mid_str}_sm'


def spacy_load(lang) -> spacy.Language:
    lib_name = get_spacy_lib(lang)
    try:
        nlp = spacy.load(lib_name)
    except (IOError, ImportError, OSError):
        logger.warning(f'Spacy model {lib_name} missed, downloading')
        spacy.cli.download(lib_name)
        nlp = spacy.load(lib_name)

    return nlp


def get_similarity(text1, text2):
    lang1 = detect_lang(text1)
    lang2 = detect_lang(text2)

    if lang1 != lang2:
        raise ValueError(f'language of "{text1}" ({lang1}) is not the same as "{text2}" ({lang2})')

    nlp = spacy_load(lang1)

    doc1 = nlp(text1)
    doc2 = nlp(text2)

    return doc1.similarity(doc2)


def merge_subtitle(video_path, subtitle_path, output_path):
    def convert2ffmpeg_path(path):
        abs_path = str(Path(path).absolute()).replace("\\", "/")

        abs_path = abs_path[0] + '\\\\' + abs_path[1:]

        return abs_path

    # check ffmpeg
    try:
        subprocess.check_output('ffmpeg -version', shell=True)
    except FileNotFoundError:
        raise RuntimeError('ffmpeg is not installed. Please install ffmpeg first.')

    subtitle_path = convert2ffmpeg_path(subtitle_path)
    style = 'FontSize=24,Bold=1'
    subprocess.call(
        f'ffmpeg -y -i "{video_path}" -vf subtitles="{subtitle_path}":force_style=\'{style}\' "{output_path}"',
        shell=True)

    logger.info(f'Subtitled video saved to {output_path}')



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import streamlit as st

context_msg = """## Context 📝

Utilize the available context to enhance the quality of your translation.
Save them as `context.yaml` in the same directory as your audio file.

> [!NOTE]
> The improvement of translation quality from Context is **NOT** guaranteed.

```yaml
background: "This is a multi-line background.
This is a basic example."
audio_type: Movie
description_map: {
  movie_name1 (without extension): "This
  is a multi-line description for movie1.",
  movie_name2 (without extension): "This
  is a multi-line description for movie2.",
  movie_name3 (without extension): "This is a single-line description for movie 3.",
}
```"""
st.write(context_msg)



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.
from pathlib import Path
from zipfile import ZipFile


def get_asr_options(beam_size, best_of, patience, length_penalty, repetition_penalty, no_repeat_ngram_size, temperature,
                    compression_ratio_threshold, log_prob_threshold, no_speech_threshold, condition_on_previous_text,
                    initial_prompt, prefix, suppress_blank, suppress_tokens, without_timestamps, max_initial_timestamp,
                    word_timestamps, prepend_punctuations, append_punctuations, hallucination_silence_threshold):
    options = {
        "beam_size": beam_size,
        "best_of": best_of,
        "patience": patience,
        "length_penalty": length_penalty,
        "repetition_penalty": repetition_penalty,
        "no_repeat_ngram_size": no_repeat_ngram_size,
        "temperature": temperature,
        "compression_ratio_threshold": compression_ratio_threshold,
        "log_prob_threshold": log_prob_threshold,
        "no_speech_threshold": no_speech_threshold,
        "condition_on_previous_text": condition_on_previous_text,
        "initial_prompt": initial_prompt,
        "prefix": prefix,
        "suppress_blank": suppress_blank,
        "suppress_tokens": [int(x) for x in suppress_tokens.split(",") if x.strip().lstrip('-').isdigit()],
        "without_timestamps": without_timestamps,
        "max_initial_timestamp": max_initial_timestamp,
        "word_timestamps": word_timestamps,
        "prepend_punctuations": prepend_punctuations,
        "append_punctuations": append_punctuations,
        "hallucination_silence_threshold": hallucination_silence_threshold,
    }

    return options


def get_vad_options(threshold, min_speech_duration_ms, max_speech_duration_s, min_silence_duration_ms,
                    window_size_samples, speech_pad_ms):
    options = {
        "threshold": threshold,
        "min_speech_duration_ms": min_speech_duration_ms,
        "max_speech_duration_s": max_speech_duration_s,
        "min_silence_duration_ms": min_silence_duration_ms,
        "window_size_samples": window_size_samples,
        "speech_pad_ms": speech_pad_ms,
    }

    return options


def get_preprocess_options(atten_lim_db):
    options = {
        'atten_lim_db': atten_lim_db
    }

    return options


def zip_files(file_paths, zip_filename='zipped'):
    file_paths = [Path(path) for path in file_paths]
    zip_filename = file_paths[0].parent.with_name(zip_filename).with_suffix('.zip')
    with ZipFile(zip_filename, 'w') as zip_object:
        _ = [zip_object.write(lrc_path, arcname=lrc_path.name) for lrc_path in file_paths]

    return zip_filename



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.
from typing import Union

from anthropic.types import Message
from openai.types.chat import ChatCompletion


class SameLanguageException(Exception):
    """
    Raised when the source language and target language are the same.
    """

    def __init__(self):
        super().__init__(
            'Source language and target language are the same, no need to translate. '
            'If you want to translate, set force_translate=True.'
        )


class ChatBotException(Exception):
    """
    Raised when chatbot fails to generate response.
    """

    def __init__(self, message):
        super().__init__(message)


class LengthExceedException(ChatBotException):
    """
    Raised when the length of generated response exceeds the limit.
    """

    def __init__(self, response: Union[ChatCompletion, Message]):
        if isinstance(response, ChatCompletion):
            prompt_tokens = response.usage.prompt_tokens
            completion_tokens = response.usage.completion_tokens
            total_tokens = response.usage.total_tokens
        elif isinstance(response, Message):
            prompt_tokens = response.usage.input_tokens
            completion_tokens = response.usage.output_tokens
            total_tokens = prompt_tokens + completion_tokens
        else:
            raise ValueError(f'Invalid response type: {type(response)}')

        super().__init__(
            f'Failed to get completion. Exceed max token length. '
            f'Prompt tokens: {prompt_tokens}, '
            f'Completion tokens: {completion_tokens}, '
            f'Total tokens: {total_tokens} '
            f'Reduce chunk_size may help.'
        )


class OpenaiFailureException(Exception):
    """
    Raised when OpenAI API fails to generate response.
    """

    def __init__(self):
        super().__init__('OpenAI API failed to generate response.')


class FfmpegException(Exception):
    def __init__(self, message):
        super().__init__(message)


class TranscribeException(Exception):
    def __init__(self, message):
        super().__init__(message)


class DependencyException(Exception):
    def __init__(self, message):
        super().__init__(f'Dependency not correctly installed: {message}')



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.

import asyncio
import os
import random
import time
from typing import List, Union, Dict, Callable

import anthropic
import httpx
import openai
from anthropic import AsyncAnthropic
from anthropic._types import NOT_GIVEN
from anthropic.types import Message
from openai import AsyncClient as AsyncGPTClient
from openai.types.chat import ChatCompletion

from openlrc.exceptions import ChatBotException, LengthExceedException
from openlrc.logger import logger
from openlrc.utils import get_messages_token_number, get_text_token_number

model2chatbot = {
    'gpt-4-0125-preview': 'gptbot',
    'gpt-4-turbo-preview': 'gptbot',
    'gpt-3.5-turbo-0125': 'gptbot',
    'gpt-3.5-turbo': 'gptbot',

    'claude-3-opus-20240229': 'claudebot',
    'claude-3-sonnet-20240229': 'claudebot',
    'claude-3-haiku-20240307': 'claudebot'
}

chatbot_map = {}


def _register_chatbot(cls):
    chatbot_map[cls.__name__.lower()] = cls
    return cls


class ChatBot:
    def __init__(self, pricing, temperature=1, top_p=1, retry=8, max_async=16, fee_limit=0.05):
        self.pricing = pricing
        self._model = None

        self.temperature = temperature
        self.top_p = top_p
        self.retry = retry
        self.max_async = max_async
        self.fee_limit = fee_limit

        self.api_fees = []

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        if model not in self.pricing:
            raise ValueError(f'Invalid model {model}.')
        self._model = model

    def estimate_fee(self, messages: List[Dict]):
        """
        Estimate the total fee for the given messages.
        """
        token_map = {'system': 0, 'user': 0, 'assistant': 0}
        for message in messages:
            token_map[message['role']] += get_text_token_number(message['content'])

        prompt_price, completion_price = self.pricing[self.model]

        total_price = (sum(token_map.values()) * prompt_price + token_map['user'] * completion_price * 2) / 1000000

        return total_price

    def update_fee(self, response):
        raise NotImplementedError()

    def get_content(self, response):
        raise NotImplementedError()

    async def _create_achat(self, messages: List[Dict], output_checker: Callable = lambda *args, **kw: True):
        raise NotImplementedError()

    async def _amessage(self, messages_list: List[List[Dict]], output_checker: Callable = lambda *args, **kw: True):
        """
        Async send messages to the GPT chatbot.
        """

        results = await asyncio.gather(
            *(self._create_achat(message, output_checker=output_checker) for message in messages_list)
        )

        return results

    def message(self, messages_list: Union[List[Dict], List[List[Dict]]],
                output_checker: Callable = lambda *args, **kw: True):
        """
        Send chunked messages to the GPT chatbot.
        """
        assert messages_list, 'Empty message list.'

        if isinstance(messages_list[0], dict):  # convert messages List[Dict] to messages_list List[List[Dict]]
            messages_list = [messages_list]

        # Calculate the total sending token number and approximated billing fee.
        token_numbers = [get_messages_token_number(message) for message in messages_list]
        logger.info(f'Max token num: {max(token_numbers):.0f}, '
                    f'Avg token num: {sum(token_numbers) / len(token_numbers):.0f}')

        # if the approximated billing fee exceeds the limit, raise an exception.
        approximated_fee = sum([self.estimate_fee(messages) for messages in messages_list])
        logger.info(f'Approximated billing fee: {approximated_fee:.4f} USD')
        self.api_fees += [0]  # Actual fee for this translation call.
        if approximated_fee > self.fee_limit:
            raise ChatBotException(f'Approximated billing fee {approximated_fee} '
                                   f'exceeds the limit: {self.fee_limit}$.')

        try:
            results = asyncio.run(self._amessage(messages_list, output_checker=output_checker))
        except ChatBotException as e:
            logger.error(f'Failed to message with GPT. Error: {e}')
            raise e
        finally:
            logger.info(f'Translation fee for this call: {self.api_fees[-1]:.4f} USD')
            logger.info(f'Total bot translation fee: {sum(self.api_fees):.4f} USD')

        return results


@_register_chatbot
class GPTBot(ChatBot):
    def __init__(self, model='gpt-3.5-turbo-0125', temperature=1, top_p=1, retry=8, max_async=16, json_mode=False,
                 fee_limit=0.05, proxy=None, base_url_config=None):
        # Pricing for 1M tokens, info from https://openai.com/pricing
        pricing = {
            'gpt-3.5-turbo-0125': (0.5, 1.5),
            'gpt-3.5-turbo': (0.5, 1.5),
            'gpt-4-0125-preview': (10, 30),
            'gpt-4-turbo-preview': (10, 30)
        }

        super().__init__(pricing, temperature, top_p, retry, max_async, fee_limit)

        self.async_client = AsyncGPTClient(
            api_key=os.environ['OPENAI_API_KEY'],
            http_client=httpx.AsyncClient(proxies=proxy),
            base_url=base_url_config['openai'] if base_url_config else None
        )

        self.model = model
        self.temperature = temperature
        self.top_p = top_p
        self.retry = retry
        self.max_async = max_async
        self.json_mode = json_mode
        self.fee_limit = fee_limit

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.async_client.close()

    def update_fee(self, response: ChatCompletion):
        prompt_price, completion_price = self.pricing[self.model]

        prompt_tokens = response.usage.prompt_tokens
        completion_tokens = response.usage.completion_tokens

        self.api_fees[-1] += (prompt_tokens * prompt_price + completion_tokens * completion_price) / 1000000

    def get_content(self, response):
        return response.choices[0].message.content

    async def _create_achat(self, messages: List[Dict], output_checker: Callable = lambda *args, **kw: True):
        logger.debug(f'Raw content: {messages}')

        response = None
        for i in range(self.retry):
            try:
                response = await self.async_client.chat.completions.create(
                    model=self.model,
                    messages=messages,
                    temperature=self.temperature,
                    top_p=self.top_p,
                    response_format={'type': 'json_object' if self.json_mode else 'text'}
                )
                self.update_fee(response)
                if response.choices[0].finish_reason == 'length':
                    raise LengthExceedException(response)
                if not output_checker(messages, response.choices[0].message.content):
                    logger.warning(f'Invalid response format. Retry num: {i + 1}.')
                    continue

                break
            except openai.RateLimitError:
                sleep_time = random.randint(30, 60)
                logger.warning(f'Rate limit exceeded. Wait {sleep_time}s before retry. Retry num: {i + 1}.')
                time.sleep(sleep_time)
            except openai.APITimeoutError:
                logger.warning(f'Timeout. Wait 3 before retry. Retry num: {i + 1}.')
                time.sleep(3)
            except openai.APIConnectionError:
                logger.warning(f'API connection error. Wait 15s before retry. Retry num: {i + 1}.')
                time.sleep(15)
            except openai.APIError:
                logger.warning(f'API error. Wait 15s before retry. Retry num: {i + 1}.')
                time.sleep(15)

        if not response:
            raise ChatBotException('Failed to create a chat.')

        return response


@_register_chatbot
class ClaudeBot(ChatBot):
    def __init__(self, model='claude-3-sonnet-20240229', temperature=1, top_p=1, retry=8, max_async=16, fee_limit=0.2,
                 proxy=None, base_url_config=None):
        # Pricing for 1M tokens, info from https://docs.anthropic.com/claude/docs/models-overview#model-comparison
        pricing = {
            'claude-3-opus-20240229': (15, 75),
            'claude-3-sonnet-20240229': (3, 15),
            'claude-3-haiku-20240307': (0.25, 1.25)
        }

        super().__init__(pricing, temperature, top_p, retry, max_async, fee_limit)

        self.async_client = AsyncAnthropic(
            api_key=os.environ['ANTHROPIC_API_KEY'],
            http_client=httpx.AsyncClient(
                proxies=proxy
            ),
            base_url=base_url_config['anthropic'] if base_url_config else None
        )

        self.model = model
        self.retry = retry
        self.max_async = max_async
        self.fee_limit = fee_limit

    def update_fee(self, response: Message):
        prompt_price, completion_price = self.pricing[self.model]

        prompt_tokens = response.usage.input_tokens
        completion_tokens = response.usage.output_tokens

        self.api_fees[-1] += (prompt_tokens * prompt_price + completion_tokens * completion_price) / 1000000

    def get_content(self, response):
        return response.content[0].text

    async def _create_achat(self, messages: List[Dict], output_checker: Callable = lambda *args, **kw: True):
        logger.debug(f'Raw content: {messages}')

        # Move "system" role into the parameters
        system_msg = NOT_GIVEN
        if messages[0]['role'] == 'system':
            system_msg = messages.pop(0)['content']

        response = None
        for i in range(self.retry):
            try:
                response = await self.async_client.messages.create(
                    max_tokens=4096,
                    model=self.model,
                    messages=messages,
                    system=system_msg,
                    temperature=self.temperature,
                    top_p=self.top_p,
                )
                self.update_fee(response)

                if response.stop_reason == 'max_tokens':
                    raise LengthExceedException(response)
                if not output_checker(messages, response.content[-1].text):
                    logger.warning(f'Invalid response format. Retry num: {i + 1}.')
                    continue

                break
            except anthropic.RateLimitError:
                sleep_time = random.randint(30, 60)
                logger.warning(f'Rate limit exceeded. Wait {sleep_time}s before retry. Retry num: {i + 1}.')
                time.sleep(sleep_time)
            except anthropic.APITimeoutError:
                logger.warning(f'Timeout. Wait 3 before retry. Retry num: {i + 1}.')
                time.sleep(3)
            except anthropic.APIConnectionError:
                logger.warning(f'API connection error. Wait 15s before retry. Retry num: {i + 1}.')
                time.sleep(15)
            except anthropic.APIError as e:
                logger.warning(f'API error: {e}. Wait 15s before retry. Retry num: {i + 1}.')
                time.sleep(15)

        if not response:
            raise ChatBotException('Failed to create a chat.')

        return response



import logging

from colorlog import ColoredFormatter

handler = logging.StreamHandler()

formatter = ColoredFormatter(
    "%(log_color)s [%(asctime)s] %(levelname)-8s [%(threadName)s] %(message)s",
    datefmt='%Y-%m-%d %H:%M:%S',
    reset=True,
    log_colors={
        'DEBUG': 'cyan',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'red,bg_white',
    },
    secondary_log_colors={},
    style='%'
)

handler.setFormatter(formatter)
logger = logging.getLogger()
logger.addHandler(handler)

logger.setLevel('INFO')



#  Copyright (C) 2024. Hao Zheng
#  All rights reserved.
import os
import shutil
import tempfile
from pathlib import Path

import streamlit as st
from st_pages import Page, show_pages
from streamlit_extras.bottom_container import bottom
from streamlit_extras.mention import mention

from openlrc import LRCer
from openlrc.gui_streamlit.utils import get_asr_options, get_vad_options, get_preprocess_options, zip_files

show_pages([
    Page("home.py", "Home", "🏠"),
])

with bottom():
    pass

# UI Title
"# openlrc 🎙📄"

'*Transcribe and translate voice into LRC file using Whisper and LLMs (GPT, Claude, et,al).*'

mention(
    label="zh-plus/openlrc",
    icon="github",  # GitHub is also featured!
    url="https://github.com/zh-plus/openlrc",
)

# Sidebar - Normal Configuration
st.sidebar.header('Configuration')

with st.sidebar.popover("API Keys"):
    openai_api_key = st.text_input("OpenAI API Key", value=os.environ.get('OPENAI_API_KEY'), key='openai_api')
    anthropic_api_key = st.text_input("Anthropic API Key", value=os.environ.get('ANTHROPIC_API_KEY'),
                                      key='anthropic_api')

whisper_model = st.sidebar.selectbox('Whisper Model',
                                     ['large-v3', 'medium', 'medium.en', 'small', 'small.en', 'base', 'base.en', 'tiny',
                                      'tiny.en'], index=0, key='whisper_model')
compute_type = st.sidebar.selectbox('Compute Type', ['int8', 'int8_float16', 'int16', 'float16', 'float32'], index=3
                                    , key='compute_type')

if not openai_api_key and not anthropic_api_key:
    chatbot_model = st.sidebar.selectbox('Chatbot Model', [], disabled=True, help="Please provide an API key first.",
                                         key='chatbot_model')
elif openai_api_key and not anthropic_api_key:
    chatbot_model = st.sidebar.selectbox('Chatbot Model',
                                         ['gpt-3.5-turbo', 'gpt-4-0125-preview', 'gpt-4-turbo-preview'], index=0,
                                         help="Model for translation. Check [pricing](/pricing) for more details.",
                                         key='chatbot_model')
elif not openai_api_key and anthropic_api_key:
    chatbot_model = st.sidebar.selectbox('Chatbot Model', ['claude-3-haiku-20240307', 'claude-3-sonnet-20240229',
                                                           'claude-3-opus-20240229'], index=0,
                                         help="Model for translation. Check [pricing](/pricing) for more details.",
                                         key='chatbot_model')
else:
    chatbot_model = st.sidebar.selectbox('Chatbot Model',
                                         ['gpt-3.5-turbo', 'gpt-4-0125-preview', 'gpt-4-turbo-preview',
                                          'claude-3-haiku-20240307', 'claude-3-sonnet-20240229',
                                          'claude-3-opus-20240229'],
                                         index=0,
                                         help="Model for translation. Check [pricing](/pricing) for more details.",
                                         key='chatbot_model')

# fee_limit = st.sidebar.number_input('Fee Limit', min_value=0.0, value=0.1, step=0.01)
fee_limit = st.sidebar.slider('Fee Limit (USD)', min_value=0.0, max_value=1.0, value=0.1, step=0.01, key='fee_limit')
consumer_thread = st.sidebar.slider('Consumer Thread', min_value=1, max_value=12, value=4, step=1,
                                    key='consumer_thread')
proxy = st.sidebar.text_input('Proxy', help='e.g.: http://127.0.0.1:7890', key='proxy')

# Sidebar: Advanced Configuration
with st.sidebar.expander("Advanced Configuration", expanded=False):
    st.write("### ASR Options")
    # ASR Options with help messages
    beam_size = st.number_input("Beam Size", value=3, min_value=1, help="Sets the beam size for the ASR.")
    best_of = st.number_input("Best of", value=5, min_value=1, help="Chooses the best result from N beams.")
    patience = st.number_input("Patience", value=1, min_value=0,
                               help="Number of times to retry ASR for better accuracy.")
    length_penalty = st.number_input("Length Penalty", value=1, min_value=0,
                                     help="Penalty for longer sequences to control output length.")
    repetition_penalty = st.number_input("Repetition Penalty", value=1.0, min_value=0.0,
                                         help="Penalty for repeated tokens to encourage diversity.")
    no_repeat_ngram_size = st.number_input("No Repeat Ngram Size", value=0, min_value=0,
                                           help="Size of ngram that cannot be repeated in the output.")
    temperature = st.number_input("Temperature", value=0.0, min_value=0.0, max_value=1.0, step=0.01,
                                  help="Controls randomness in beam search. Lower values make results more deterministic.")
    compression_ratio_threshold = st.number_input("Compression Ratio Threshold", value=None, min_value=0.0,
                                                  help="Threshold for compression ratio.")
    log_prob_threshold = st.number_input("Log Prob Threshold", value=None, min_value=0.0,
                                         help="Threshold for log probability.")
    no_speech_threshold = st.number_input("No Speech Threshold", value=None, min_value=0.0,
                                          help="Threshold for no speech.")
    condition_on_previous_text = st.checkbox("Condition on Previous Text", value=False,
                                             help="The previous output of the model will be provided as a prompt for the next window. (May enhance model hallucinations)")
    initial_prompt = st.text_input("Initial Prompt", value=None, help="Initial prompt for the ASR.")
    prefix = st.text_input("Prefix", value=None, help="Prefix for the ASR.")
    suppress_blank = st.checkbox("Suppress Blank", value=True, help="Whether to suppress blank tokens in the output.")
    suppress_tokens = st.text_input("Suppress Tokens", value="-1",
                                    help="Tokens to be suppressed in the output, int values split by ','.")
    without_timestamps = st.checkbox("Without Timestamps", value=False, help="Generate output without timestamps.")
    max_initial_timestamp = st.number_input("Max Initial Timestamp", min_value=0.0, value=0.0, step=0.01,
                                            help="Maximum initial timestamp for the output.")
    word_timestamps = st.checkbox("Word Timestamps", value=True,
                                  help="Whether to include timestamps for each word in the output.")
    prepend_punctuations = st.text_input("Prepend Punctuations", value="\"'“¿([{-",
                                         help="Punctuations to prepend to the output.")
    append_punctuations = st.text_input("Append Punctuations", value="\"'.。,，!！?？:：”)]}、",
                                        help="Punctuations to append to the output.")
    hallucination_silence_threshold = st.number_input("Hallucination Silence Threshold", min_value=0, value=2,
                                                      help="Threshold for determining hallucinations in silence periods.")

    # VAD Options with help messages
    st.write("### VAD Options")
    threshold = st.slider("Threshold", min_value=0.0, max_value=1.0, value=0.382,
                          help="Threshold for voice activity detection.")
    min_speech_duration_ms = st.number_input("Min Speech Duration (ms)", min_value=0, value=250,
                                             help="Minimum duration of speech for it to be considered valid.")
    max_speech_duration_s = st.number_input("Max Speech Duration (s)", min_value=0.0, value=1e+308, format="%e",
                                            help="Maximum duration of speech. Use a high value to indicate no limit.")
    min_silence_duration_ms = st.number_input("Min Silence Duration (ms)", min_value=0, value=2000,
                                              help="Minimum duration of silence for splitting phrases.")
    window_size_samples = st.number_input("Window Size Samples", min_value=1, value=1024,
                                          help="Size of the analysis window for VAD.")
    speech_pad_ms = st.number_input("Speech Pad (ms)", min_value=0, value=400, help="Padding added to speech segments.")

    # Preprocess Options with help messages
    st.write("### Preprocess Options")
    atten_lim_db = st.number_input("Atten Lim DB", value=15, min_value=0, help="Limit for attenuation in decibels.")

with st.form("transcribe_translate_form"):
    st.write("## Transcribe and Translate")

    # File upload (assuming 'paths' will be handled separately or is part of the form)
    files = st.file_uploader("Upload files", accept_multiple_files=True,
                             type=['mp3', 'wav', 'flac', 'm4a', 'mp4', 'avi', 'mkv', 'webm', 'mov', 'wmv', 'flv', ])

    # Currently, st.file_uploader cant return paths. (On planning in https://roadmap.streamlit.app/#future)
    # Thus, save them into temporary files and use lrcer.run on them
    tmpdir = tempfile.mkdtemp()  # Remember to delete it when the process is done
    paths = []
    for file in files:
        with open(os.path.join(tmpdir, file.name), 'wb') as f:
            f.write(file.read())
        paths.append(os.path.join(tmpdir, file.name))

    src_lang = st.selectbox("Source Language",
                            options=['Auto Detect', 'ca', 'zh', 'hr', 'da', 'nl', 'en', 'fi', 'fr', 'de', 'el', 'it',
                                     'ja', 'ko', 'lt', 'mk', 'nb', 'pl', 'pt', 'ro', 'ru', 'sl', 'es', 'sv', 'uk'],
                            index=0,
                            format_func=lambda x: 'Auto Detect' if x == 'Auto Detect' else x.upper(),
                            help='Currently bottleneck-ed by Spacy')
    target_lang = st.text_input("Target Language", value='zh-cn', help='Language code for translation target')
    prompter = st.selectbox("Prompter", options=['base'], disabled=True, help='Currently, only `base` is supported.')
    context_path = st.text_input("Context Path",
                                 help='Additional context to aid translation. Check [context](/context) for more details. ')

    col1, col2, col3 = st.columns(3)
    with col1:
        skip_trans = st.checkbox("Skip Translation")
    with col2:
        noise_suppress = st.checkbox("Noise Suppression")
    with col3:
        bilingual_sub = st.checkbox("Bilingual Subtitles")

    # Form submission button
    submitted = st.form_submit_button("GO!", type='primary')

if submitted:
    # Assuming 'paths' is defined or obtained elsewhere in your app
    src_lang = None if src_lang == 'Auto Detect' else src_lang

    with st.spinner('Running...'):
        lrcer = LRCer(whisper_model=whisper_model, compute_type=compute_type, chatbot_model=chatbot_model,
                      fee_limit=fee_limit, consumer_thread=consumer_thread,
                      asr_options=get_asr_options(
                          beam_size, best_of, patience, length_penalty, repetition_penalty, no_repeat_ngram_size,
                          temperature, compression_ratio_threshold, log_prob_threshold, no_speech_threshold,
                          condition_on_previous_text, initial_prompt, prefix, suppress_blank, suppress_tokens,
                          without_timestamps, max_initial_timestamp, word_timestamps, prepend_punctuations,
                          append_punctuations, hallucination_silence_threshold),
                      vad_options=get_vad_options(
                          threshold, min_speech_duration_ms, max_speech_duration_s, min_silence_duration_ms,
                          window_size_samples, speech_pad_ms),
                      preprocess_options=get_preprocess_options(atten_lim_db),
                      proxy=proxy, )
        results = lrcer.run(paths, src_lang=src_lang, target_lang=target_lang, prompter=prompter,
                            context_path=context_path if context_path else None,
                            skip_trans=skip_trans, noise_suppress=noise_suppress, bilingual_sub=bilingual_sub)
    print(paths)
    print(results)

    result_file_path = results[0] if len(results) == 1 else zip_files(results)

    print(result_file_path)

    with open(result_file_path, 'rb') as f:
        st.download_button("Download LRC Files", f, file_name=Path(result_file_path).name)

    # Remove tmpdir contents
    shutil.rmtree(tmpdir, ignore_errors=True)

    print(f'Removed {tmpdir}.')
